/* ---------------------------------------------------------------------------
**	Texas Instruments TVP 3026 PLL driver
** ---------------------------------------------------------------------------
**
**	Copyright	1997		Jonas Borgstrom
**			1997		Steffen Seeger
**
**	This program is free software; you can redistribute it and/or modify
**	it under the terms of the GNU General Public License as published by
**	the Free Software Foundation; either version 2, or (at your option)
**	any later version.
**
**	You should have received a copy of the GNU General Public License
**	along with this program; see the file COPYING.  If not, write to
**	the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
** ---------------------------------------------------------------------------
**	This code was derived from the following sources of information:
**
**	[1]	"TVP 3026 Data Manual SLAS098B", (C) Texas Instruments 1996
*/

#include <ggi/maintainers.h>
#define	MAINTAINER	Jonas_Borgstrom
#define	DRIVER_NAME	"TVP 3026 PLL driver"
#define	DRIVER_REV	"$Revision: 1.8 $"

#define	DEBUG_LEVEL	0

#include <ggi/module.h>
#include "pll.h"

#include "TI/tvp3026.h"

static struct kgi_clock clock =
{
	"Texas Instruments",		/* manufacturer		*/
	"TVP 3026",			/* model		*/
	GM_ALL,				/* supported modes 	*/
	0,				/* special capabilities	*/
	{ 0x7FFF, 0x7FFF },		/* maximum resolution	*/

	CLK_TYPE_PROG,			/* clock type		*/
	{{
		13800000, 250000000,	/* clock range		*/
	}}
};

static const struct pll_limits dclk_pll = 
{
	14318180,			/* fref				*/
	{ 8, 1 },			/* factor		 	*/
	{ 110000000, 250000000 },	/* fvco range			*/
	{ 0, 3 },			/* post-VCO divider range	*/
	{ 3, 64 },			/* dividend range		*/
	{ 3, 25 }			/* divisor range		*/
};

#define	INVALID_COMMAND	((ggi_uint) -1)

static ggi_uint pll_command(struct pll_state *s, struct kgi_mode *mode)
{
	ASSERT(s->p   >= dclk_pll.p.min    &&  s->p   <= dclk_pll.mul.max);
	ASSERT(s->mul >= dclk_pll.mul.min  &&  s->mul <= dclk_pll.mul.max);
	ASSERT(s->div >= dclk_pll.div.min  &&  s->div <= dclk_pll.div.max);

	return (65 - s->mul) | ((65 - s->div) << 6) | (s->p << 12);
}

static inline void EXT_out8(struct kgi_display *dpy, uint8 val, uint8 reg)
{
	DAC_out8(dpy, reg, DAC_EXT_ADDR);
	DAC_out8(dpy, val, DAC_EXT_DATA);
}

static void pll_set(struct kgi_display *dpy, ggi_uint clock_cmd)
{
	ggi_uint pp, pm, pn;

	/*	extract PCLK and LCLK settings from clock_cmd.
	*/
	pm = clock_cmd & 0x3F;
	pn = (clock_cmd >> 6) & 0x3F;
	pp = (clock_cmd >> 12) & 0x03;

	DEBUG2("pclk m = %.2x, n = %.2x, p = %.2x, cmd = %i",
		pm, pn, pp, clock_cmd);

	/*	select PCLK PLL as clock source
	*/
	DAC_out8(dpy, DAC_EXT_ClockSelect, DAC_EXT_ADDR);
	DAC_out8(dpy, (DAC_in8(dpy, DAC_EXT_DATA) & ~DAC_EXT1A_ClkMask) |
		DAC_EXT1A_ClkPClkPLL, DAC_EXT_DATA);

	/*	disable PCLK PLL
	*/
	EXT_out8(dpy, (DAC_PLL_P << DAC_EXT2C_PClkShift) |
		(DAC_PLL_P << DAC_EXT2C_LClkShift), DAC_EXT_PLLAddress);
	EXT_out8(dpy, 0, DAC_EXT_PClkData);

	/*	select PLL on CS lines
	*/
	CLK_out8(dpy, 3, 0);

	/*	set PCLK N, M and P values and wait until PCLK PLL is locked.
	*/
	EXT_out8(dpy, 0, DAC_EXT_PLLAddress);
	DAC_out8(dpy, DAC_EXT_PClkData, DAC_EXT_ADDR);

	DAC_out8(dpy, pn | DAC_PClkN_Bias, DAC_EXT_DATA);
	DAC_out8(dpy, pm | DAC_PClkM_Bias, DAC_EXT_DATA);
	DAC_out8(dpy, pp | DAC_PClkP_Bias | 
		DAC_PClkP_PClkEnable | DAC_PClkP_PLLEnable, DAC_EXT_DATA);
	
	/*	LCLK is handled by the DAC driver.
	*/
}

/*
**	KGI module interface
*/

#define	NO_KGIM_CLOCK_INIT
#define	NO_KGIM_CLOCK_DONE
#define	NO_KGIM_CLOCK_CHECK_MODE
#include "pll.inc"

ggi_sint kgim_clock_init(struct kgi_card *card, struct kgi_display *dpy)
{
	NOTICE(DRIVER_NAME " rev " DRIVER_REV);

	DAC_out8(dpy, DAC_EXT3F_ID, DAC_EXT_ADDR);
	if (DAC_in8(dpy, DAC_EXT_DATA) != DAC_EXT3F_TVP3026_ID) {

		NOTICE("detection failed");
		return -E(CLOCK, UNKNOWN);
	}

	DAC_out8(dpy, DAC_REV, DAC_EXT_ADDR);
	DEBUG1("TVP 3026 rev %.2x detected.", DAC_in8(dpy, DAC_EXT_DATA));

	card->clock = &clock;
	return EOK;
}

void kgim_clock_done(struct kgi_display *dpy)
{
	dpy->card->clock = NULL;
	NOTICE(DRIVER_NAME " removed");
}

#define	MIN_FREQ clock.clock.range[0].min
#define	MAX_FREQ clock.clock.range[0].max

ggi_sint kgim_clock_check_mode(struct kgi_display *dpy, struct kgi_mode *mode, 
	enum kgi_tm_cmd cmd)
{
	struct pll_state state;
	/*	We always have to provide DCLK for internal reference.
	*/
	ggi_uint newfreq = (cmd == CMD_PROPOSE) ? 0 : mode->dclk;

	switch (cmd) {

	case CMD_PROPOSE:
		mode->clock = INVALID_COMMAND;
		return EOK;

	case CMD_LOWER:

		if (newfreq < MIN_FREQ) {

			DEBUG1("DCLK = %i Hz, CLK = %i Hz is too low",\
				mode->dclk, newfreq);
			return -E(CLOCK, UNKNOWN);
		}

		if (newfreq > MAX_FREQ) {

			newfreq = MAX_FREQ;
		}

		if (pll_best_state(newfreq, cmd, &state, &dclk_pll)) {

			DEBUG1("invalid request for DCLK = %i Hz, CLK = %i Hz",\
				mode->dclk, newfreq);
			return -E(CLOCK, UNKNOWN);
		}

		mode->dclk = state.f;
		return EOK;

	case CMD_RAISE:

		if (MAX_FREQ < newfreq) {

			DEBUG1("DCLK = %i Hz, CLK = %i Hz is too high", \
				mode->dclk, newfreq);
			return -E(CLOCK, UNKNOWN);
		}

		if (newfreq < MIN_FREQ) {

			newfreq = MIN_FREQ;
		}

		if (pll_best_state(newfreq, cmd, &state, &dclk_pll)) {

			DEBUG1("invalid request for DCLK = %i Hz, CLK = %i Hz",\
				mode->dclk, newfreq);
			return -E(CLOCK, UNKNOWN);
		}

		mode->dclk = state.f;
		return EOK;

	case CMD_CHECK:

		mode->clock = INVALID_COMMAND;
		if (pll_best_state(newfreq, CMD_LOWER, &state, &dclk_pll)) {

			DEBUG1("invalid request for DCLK = %i Hz, CLK = %i Hz",\
				mode->dclk, newfreq);
			return -E(CLOCK, UNKNOWN);
		}

		mode->clock = pll_command(&state, mode);
		return EOK;

	default:
		INTERNAL_ERROR;
		return -E(CLOCK, UNKNOWN);
	}
}
