/* --------------------------------------------------------------------------
**	GGI event definitions
** --------------------------------------------------------------------------
**
**	Copyright (C) 1995-1997 by Steffen Seeger.
**
**	This program is free software; you can redistribute it and/or modify
**	it under the terms of the GNU General Public License as published by
**	the Free Software Foundation; either version 2, or (at your option)
**	any later version.
**
**	You should have received a copy of the GNU General Public License
**	along with this program; see the file COPYING.  If not, write to
**	the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
**
** ---------------------------------------------------------------------------
**	MAINTAINER	Steffen_Seeger
*/

#ifndef _GGI_EVENTS_H
#define	_GGI_EVENTS_H

#include <ggi/keyboard.h>
#include <sys/errno.h>
#include <string.h> /* Needed for memcpy() on OSF1 */

typedef enum ggi_input_type
{
	GGI_IC_NULL = 0,

	GGI_IC_KEYBOARD = 0,		/* character input devices	*/
		GGI_IT_KEYBOARD,	/* a simple keyboard		*/
		GGI_IT_VOICE,		/* voice recognition 		*/

	GGI_IC_2DPOINTER = 0x10000000,	/* simple 2D pointing devices	*/
		GGI_IT_MOUSE,		/* all kinds of mice		*/
		GGI_IT_TRACKBALL,	/* a mouse on it's back ;-)	*/
		GGI_IT_DIGITIZER,	/* digitizer boards with pen	*/
		GGI_IT_LIGHTPEN,	/* are there any left?		*/
		GGI_IT_JOYSTICK,	/* the thing used for gaming	*/

	GGI_IC_3DPOINTER = 0x20000000,	/* 3D pointing devices	 	*/
		GGI_IT_3DMOUSE,

	GGI_IC_CONTROLLER = 0x30000000,	/* any other stuff		*/
		GGI_IT_DATAGLOVE,
		GGI_IT_VR_HEAD,

	GGI_IC_KEYPAD,			/* character input but no kbd	*/

	GGI_IC_MASK = (int)0xF0000000	/* mask to get class		*/

} ggi_input_type;

typedef enum ggi_event_type {

	evNothing = 0,	/* event is not valid. (must be zero)	*/
	evCommand,	/* report command/do action		*/
	evBroadcast,	/* notification of general interest	*/
	evDeviceInfo,	/* report input device information	*/ 

	evRawData,	/* raw data received from device	*/

	evKeyPress,	/* key has been pressed			*/
	evKeyRelease,	/* key has been released		*/
	evKeyRepeat,	/* automatically repeated keypress	*/
	evKeyState,	/* resynchronize keys state		*/

	evPtrRelative,	/* pointer movements reported relative	*/
	evPtrAbsolute,	/* pointer movements reported absolute	*/
	evPtrButtonPress,	/* pointer button pressed	*/
	evPtrButtonRelease,	/* pointer button released	*/
	evPtrState,	/* resynchronize pointer state		*/

	evValRelative,	/* valuator change (reported relative)	*/
	evValAbsolute,	/* valuator change (reported absolute)	*/
	evValState,	/* resynchronize valuator state		*/

	evLast		/* must be less than 33			*/

}  ggi_event_type;

#define	EVMASK(x)	em ## x = (1 << ev ## x)
typedef enum ggi_event_mask {

	EVMASK(Nothing),
	EVMASK(Command),
	EVMASK(Broadcast),
	EVMASK(DeviceInfo),

	EVMASK(RawData),

	EVMASK(KeyPress),
	EVMASK(KeyRelease),
	EVMASK(KeyRepeat),
	EVMASK(KeyState),
	emKey		= emKeyPress | emKeyRelease | emKeyRepeat,
	emKeyboard	= emKey | emKeyState,

	EVMASK(PtrRelative),
	EVMASK(PtrAbsolute),
	EVMASK(PtrButtonPress),
	EVMASK(PtrButtonRelease),
	EVMASK(PtrState),
	emPtrMove	= emPtrRelative | emPtrAbsolute,
	emPtrButton	= emPtrButtonPress | emPtrButtonRelease,
	emPointer	= emPtrMove | emPtrButton | emPtrState,

	EVMASK(ValRelative),
	EVMASK(ValAbsolute),
	EVMASK(ValState),
	emValuator	= emValRelative | emValAbsolute | emValState,

	emAll	= ((1 << evLast) - 1) & ~emNothing

} ggi_event_mask;
#undef EVMASK


/*	Symbolic names for the keyboard LEDs. A keyboard driver should just
**	set the ones that are possible and ignore the rest.
*/
#define	KBD_LED_SCROLLOCK	0
#define	KBD_LED_NUMLOCK		1
#define	KBD_LED_CAPSLOCK	2

/*	We allow for up to eight modifier keys, the names below have no
**	meaning at all, but it is recommended to map them according to
**	their names. They are unlikely to be extended, just because
**	you normally have only ten fingers.
*/
typedef enum ggi_modifier_keys {

	mkShift,	mkCtrl,		mkAlt,		mkAltGr,
	mkShiftL,	mkShiftR,	mkCtrlL,	mkCtrlR

} ggi_modifier_keys;

#define	MAX_NR_KEYMAPS		256
#define	MAX_NR_USER_KEYMAPS	MAX_NR_KEYMAPS

/*
** NOTE	We depend on MODIFIER_DATA being declared _directly_ after COMMON_DATA.
*/
#define	MODIFIER_DATA\
	uint16	effect;		/* current modifiers in effect	*/\
	uint16	normal;		/* normal modifiers		*/\
	uint16	locked;		/* locked modifiers		*/\
	uint16	sticky		/* sticky modifiers		*/

#define	COMMON_DATA\
	uint8	size;		/* size of event in bytes	*/\
	uint8	type;		/* type of this event		*/\
	uint8	focus;		/* focus this is reported from	*/\
	uint8	device;		/* who sent this		*/\
	uint32	time		/* timestamp			*/

/*	this information is reported with all events. Use the <any> field
**	in a ggi_event structure to access these fields.
*/ 
typedef struct ggi_any_event {

	COMMON_DATA;

} ggi_any_event;

/*	These are used inernally either to the application or the
**	kernel. The recipient must not store references to the data.
**	If the data information is needed afterwards, copy it!
*/
typedef struct ggi_cmd_event {

	COMMON_DATA;

	uint32	code;		/* command/request code		*/
	char	data[1];	/* command related data		*/

} ggi_cmd_event;

/*	This reports a block of raw data received from the device.
**	Maximum number of byte is limited so that the event struct is not more
**	than 256 bytes.
*/
typedef struct ggi_raw_event {

	COMMON_DATA;

	uint8	data[255 - sizeof(ggi_any_event)];

} ggi_raw_event;

/*	key events should be used to report events obtained from keys and
**	other switches. 
*/
typedef struct ggi_key_event {

	COMMON_DATA;

	MODIFIER_DATA;
	uint32	sym;		/* action(effective, keycode)	*/
	uint32	code;		/* key pressed			*/

} ggi_key_event;

/*	This is used to report change of pointer position. 
**	Depending on the event type, the values are either absolute
**	or relative.
*/
typedef struct ggi_pmove_event {

	COMMON_DATA;

	MODIFIER_DATA;		/* current state of modfiers	*/
	sint32	x, y;		/* absolute/relative position	*/
	sint32	wheel;		/* absolute/relative wheel	*/

	uint32	dummy[2];	/* reserved (future extensions)	*/ 

} ggi_pmove_event;

/*	Button events are sent to report a change in pointer button
**	state: 
**	Bit#0 = primary button (usually left)
**	Bit#1 = secondary button (usually right)
**	Bit#2 = tertiary buttin (usually middle)
**	All further button in some "logical" ordering from then on.
*/
typedef struct ggi_pbutton_event {

	COMMON_DATA;

	MODIFIER_DATA;		/* current state of modifiers	*/
	uint32	state;		/* new button state		*/
	uint32	button;		/* button(s) that caused event	*/

} ggi_pbutton_event;

/*	valuator events report a change of up to 32 of the
**	input device valuators. Only a range of 32 valuators beginning
**	from first can be reported with one event. For performance reasons,
**	only changed values are reported in value[] thus you can't use
**	(num-first) as index to get info for valuator (num).
*/
typedef struct ggi_val_event {

	COMMON_DATA;

	uint32	first;		/* first valuator reported	*/
	uint32	changed;	/* modifiers reported		*/
	uint32	value[32];	/* absolute/relative values	*/

} ggi_val_event;

typedef union ggi_event {

	uint8			size;		/* size of this event	*/

	ggi_any_event		any;		/* access COMMON_DATA	*/
	ggi_cmd_event		cmd;		/* command/broadcast	*/
	ggi_raw_event		raw;		/* raw data event	*/
	ggi_val_event		val;		/* valuator change	*/
	ggi_key_event		key;		/* key press/release	*/
	ggi_pmove_event		pmove;		/* pointer move		*/
	ggi_pbutton_event	pbutton;	/* pointer buttons	*/

} ggi_event;

typedef struct ggi_event_buffer {

	struct ggi_event_buffer *next;
	void	*priv;

	uint32	bigbang;	/* start of time		*/
	ggi_uint	size;		/* size of buffer (pwr of 2)!	*/
	ggi_uint	read;		/* current read position	*/
	ggi_uint	write;		/* current write position	*/

	ggi_uint	overflow;	/* overflow flag		*/
	void	*buf;		/* circular buffer		*/

} ggi_event_buffer;

#ifdef __GNUC__	/* suppress those nasty warnings */
#define inline __inline__
#endif

#if (defined(__sgi) || defined(__osf__)) && !defined(__GNUC__)/* { */
extern int  ggiEventPut(ggi_event_buffer *buf, ggi_event *ev);
extern void ggiEventSkip(ggi_event_buffer *buf);
extern int  ggiEventCopy(ggi_event_buffer *buf, ggi_event *ev);

#else /* } { */

extern inline int ggiEventPut(ggi_event_buffer *buf,
	ggi_event *ev)
{
	register ggi_uint start = buf->write;
	register ggi_uint end = start + ev->size;

	ASSERT(ev->any.type != evNothing);

	if (end <= (buf->read + buf->size)) {

		buf->write = end;
		
		DEBUG2("copy event to %i-%i", start, end);

		if ((end <= buf->size) || (buf->size <= start)) {

			NOTICE2("single burst");
			memcpy((uint8 *)buf->buf + (start & (buf->size-1)), ev,
				ev->size);
		} else {

			register ggi_uint cnt = buf->size - start;

			NOTICE2("dual burst");

			memcpy((uint8 *)buf->buf + start, ev, cnt);
			memcpy(buf->buf, ((uint8 *)ev) + cnt, ev->size - cnt);
		}
		return EOK;

	} else {

		NOTICE2("event buffer overflow");
		buf->overflow = 1;
		return -ENOMEM;
	}
}

#define ggiEventAvailable(buf) (buf->read < buf->write)

extern inline void ggiEventSkip(ggi_event_buffer *buf)
{
	ggi_uint size = *((uint8 *)((uint8 *)buf->buf + buf->read));

	if (buf->read < buf->write) {

		DEBUG2("skip event from %i size %i", buf->read, size);
		buf->read += size;
		if (buf->read >= buf->size) {

			buf->read -= buf->size;
			buf->write -= buf->size;
		}
		DEBUG2("buf->read %i, buf->write %i", buf->read, buf->write);
	}
}

extern inline int ggiEventCopy(ggi_event_buffer *buf, ggi_event *ev)
{
	register ggi_uint start = buf->read;
	ggi_uint size = *((uint8 *)((uint8 *)buf->buf + start)); 
	register ggi_uint end = start + size;

	if (start < buf->write) {

		DEBUG2("copy event from %i-%i", start, end);
		
		if ((end <= buf->size) || (buf->size <= start)) {

			NOTICE2("single burst");
			memcpy(ev, (uint8 *)buf->buf + (start & (buf->size-1)), size);

		} else {

			register ggi_uint cnt = buf->size - start;

			NOTICE2("dual burst");
			memcpy(ev, (uint8 *)buf->buf + start, cnt);
			memcpy(((uint8 *)ev) + cnt, buf->buf, size - cnt);
		}
		return EOK;

	} else {

		NOTICE2("no event available");
		return -EAGAIN;	/* no event yet */
	}
}

#endif	/* #ifdef __sgi } */
#endif	/* #ifdef _GGI_EVENTS_H */
