/*
 * ----------------------------------------------------------------------------
 *	KGI manager interface definitions
 * ----------------------------------------------------------------------------
 *
 *	Copyright (C)	by Steffen Seeger
 *	1995, 1996	and Andreas Beck
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *	the Free Software Foundation; either version 2, or (at your option)
 *	any later version.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; see the file COPYING.  If not, write to
 *	the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * ----------------------------------------------------------------------------
 *	MAINTAINER	Steffen_Seeger
 */

#ifndef _GGI_KGI_H
#define _GGI_KGI_H

#ifdef __KERNEL__
#include <linux/sched.h>
#include <linux/fs.h>
#include <linux/tty.h>
#include <linux/malloc.h>
#endif

#define	NEED_KGI_PROTOTYPES
#include <ggi/ggi.h>

/* ----------------------------------------------------------------------------
 *	useful macros
 * ----------------------------------------------------------------------------
 */

#ifndef MAINTAINER
#define	MAINTAINER	"linux-ggi@eskimo.com"
#endif

#define INTERNALERROR printk("%s: %i: internal error, please report to " \
			MAINTAINER ".\n",__FILE__,__LINE__)
#define	INVALID_PARAMS	TRACE(printk(KERN_ERR "%s:%i: invalid parameter(s)",\
				__FILE__,__LINE__))

#ifdef __KERNEL__

	#define	LinuxVersionCode(v, p, s) (((v)<<16)+((p)<<8)+(s))

	/*	use these macros to allocate kernel memory. 
	**	kgi_malloc_small() may be used in IRQ environments, but may
	**	likely fail for big chunks requested there. In general it shall
	**	be used when allocating small (< PAGESIZE) chunks of memory.
	**	kgi_malloc/free_huge() are used to allocate big chunks of memory
	**	such as console backround buffers.
	*/
	#define kgi_malloc_small(size, pri)	kmalloc(size, pri)
	#define	kgi_free_small(addr)	kfree(addr)

	#define	kgi_malloc_huge(size)	vmalloc(size)
	#define	kgi_free_huge(addr)	vfree(addr)

	/*	Use the paddr_t type to store a physical address.
	**	Use the caddr_t type to store a virtual address.
	*/
	typedef unsigned long	phys_addr;

	#define NOT_TESTED \
	printk(\
    	"\nGood! You hit untested code in %s line %i. Please report whether\n"\
	"there were any difficulties or not. (Simply send this message and a\n"\
	"short explanation of your configuration to " MAINTAINER ".\n\n\n",\
	__FILE__,__LINE__)

	#ifdef DEBUG
		#define TRACE(X) X
	#else
		#define TRACE(X)
	#endif
#endif

/* ----------------------------------------------------------------------------
**	KGI module (display driver)
** ----------------------------------------------------------------------------
*/

#ifdef __KERNEL__
	#define	IS_CRITICAL	unsigned long __critical_flags 
	#define	ENTER_CRITICAL	save_flags(__critical_flags); cli()
	#define	LEAVE_CRITICAL	restore_flags(__critical_flags)
#endif

#define	KGI_COMMONVERSION	01

	enum kgi_flags
	{
		FL_MONITOR	= 0x000000FF,
		FL_MON_NORMAL	  = 0x00000000,	/* normal display	*/
		FL_MON_RIGHT_EYE  = 0x00000001,	/* stereo right screen	*/
		FL_MON_LEFT_EYE   = 0x00000002,	/* stereo left screen	*/
		FL_MON_STEREO     = 0x00000003,	/* is stereo?		*/
		FL_MON_INTERLACE  = 0x00000004,	/* interlaced operation	*/
		FL_MON_POWERSAVE  = 0x00000008,	/* has powersave option	*/
		FL_MON_PROPSIZE   = 0x00000010,	/* size depends on mode	*/

		FL_DAC		= 0x0000FF00,
		FL_DAC_8BIT	= 0x00000100,	/* 8 or 6 bit dac	*/
		FL_DAC_2XCLOCK	= 0x00000200,	/* internal 2*clk	*/
		FL_DAC_GAMMA	= 0x00000400,	/* true color LUT	*/

		FL_CLOCK	= 0x00FF0000,

		FL_CHIPSET	= 0xFF000000,
		FL_CHIP_LMAP	   = 0x01000000,/* linear mapping	*/
		FL_CHIP_DOUBLESCAN = 0x02000000,
		FL_CHIP_ACCEL	   = 0x04000000,
		FL_CHIP_COLORTEXT  = 0x08000000
	};

	#define	KGI_COMMON char manufact[64],	/* manufacturer		*/\
				model[64];	/* model/trademark	*/\
	          ggi_graphmask	modes;		/* supported modes	*/\
	         enum kgi_flags	flags;		/* special capabilities	*/\
	              ggi_coord	maxdots;	/* maximum resolution	*/

	struct	kgi_common
	{
		KGI_COMMON
	};

#define	MONITORVERSION	02

	enum	kgi_display_type
	{
		MONO_DISPLAY	= 0x00000000,	/* b&w or grayscale	*/
			HERCMONO,		/* good old Hercules	*/
			LCD,			/* mono LCD		*/
			VGA_MONO,		/* mono VGA monitor	*/

		RGB_DISPLAY	= 0x40000000,	/* RGB color display	*/
			EGA_8,			/* EGA (8 colors)	*/
			EGA_16,			/* EGA (16 colors)	*/
			CGA,			/* CGA color monitor	*/
			VGA_COLOR,		/* VGA color monitor	*/
			PLASMA,			/* plasma flat panel???	*/
			TFT,			/* (???)		*/
			COLOR_LCD,		/*Liquid Crystal Display*/

		NON_RGB_DISPLAY	= 0x80000000,	/* e.g. RGY		*/

		DISPLAY_TYPE	= 0xC0000000
	};

	enum	kgi_sync
	{
		SYNC_SEPARATE 	= 0x0000,	/* normal separate sync	*/
		SYNC_ON_RED     = 0x0001,	/* sync on r line	*/
		SYNC_ON_GREEN   = 0x0002,	/* sync on g line	*/
		SYNC_ON_BLUE    = 0x0004,	/* sync on b line	*/
		SYNC_SERRATE	= 0x0008,	/* do hsync during vsync*/
		SYNC_COMPOSITE	= 0x0010,	/* digital composite	*/
		SYNC_TYPEMASK	= 0x001F,
		SYNC_NORMAL	= SYNC_SERRATE,

		SYNC_PEDESTRAL	= 0x0020,	/* need blank pedestral	*/
		SYNC_0700_0300	= 0x0000,	/* <white>_<sync> level	*/
		SYNC_0714_0286	= 0x0040,
		SYNC_0100_0400	= 0x0080,
		SYNC_UNKNOWNLEVEL=0x00C0,	/* not defined by VESA!	*/
		SYNC_SIGNALLEVEL= 0x00C0,	/* mask for signal level*/
		
		SYNC_STANDBY	= 0x0100,	/* standby mode allowed	*/
		SYNC_SUSPEND	= 0x0200,	/* suspend mode allowed */
		SYNC_OFF	= 0x0400,	/* switch off allowed	*/
		SYNC_VESA_DPMS	= 0x0700,	/* has (full) VESA-DPMS	*/

		SYNC_SSYNC_RIGHT= 0x0000,	/* right image on SSYNC	*/
		SYNC_SSYNC_LEFT	= 0x1000,	/* left image on SSYNC	*/
		SYNC_MULTISYNC	= 0x2000,	/* multisync monitor	*/
	};

	struct kgi_monitor 
	{
		KGI_COMMON

		ggi_coord	size;		/* physical size in mm	*/
	   enum kgi_display_type disptype;	/* display type		*/
	   enum kgi_sync	synctype;	/* type of syncing	*/
		ggi_range	dclk;		/* bandwidth		*/
		ggi_range	hfreq;		/* horizontal frequency	*/
		ggi_range	vfreq;		/* vertical frequency	*/

		ggi_color	white;		/* white point		*/
		ggi_uint		gamma;		/* gamma value		*/
	};

#define	CHIPSETVERSION	01

	struct kgi_chipset
	{
		KGI_COMMON

		int		ramsize;	/* installed RAM size	*/
		ggi_range	dclk;		/* DCLK range		*/
	};

#define RAMDACVERSION	01

	enum kgi_dac_type
	{
		DAC_8BIT_PB   = 0x01,		/*   8 bit pixel bus	*/
		DAC_16BIT_PB  = 0x02,		/*  16 bit pixel bus	*/
		DAC_24BIT_PB  = 0x04,		/*  24 bit pixel bus	*/
		DAC_32BIT_PB  = 0x08,		/*  32 bit pixel bus	*/
		DAC_64BIT_PB  = 0x10,		/*  64 bit pixel bus	*/
		DAC_128BIT_PB = 0x20,		/* 128 bit pixel bus	*/
		DAC_PB_MASK   = 0x3F,		/* pixel bus width mask	*/
		DAC_VRAM      = 0x80000000	/* VRAM shift clock	*/
	};

	struct kgi_ramdac
	{
		KGI_COMMON

		enum kgi_dac_type dactype;	/* general device type	     */
		ggi_range 	normal;		/* DCLK range (normal mode)  */
		ggi_range 	dblclk;		/* DCLK (internal doubling)  */
        	ggi_range	clock[GT_LAST];	/* DCLK ranges		     */
	};

#define	CLOCKVERSION	01

	#define	MAX_FIXED_CLOCKS	16
	#define	MAX_PROG_CLOCKS		8

	enum kgi_clock_type
	{
		CLK_TYPE_FIXED,	CLK_TYPE_PROG
	};

	struct kgi_clock
	{
		KGI_COMMON

		enum kgi_clock_type	type;
		union
		{
			int		freq[MAX_FIXED_CLOCKS];
			ggi_range	range[MAX_PROG_CLOCKS];
		} clock;
	};

#define	MODEVERSION	04

	struct kgi_timing
	{			/* all parameters are in dots	*/
		ushort	width;		/* visible pixels/lines	*/
		ushort	blankstart;	/* assertion of BLANK	*/
		ushort	syncstart;	/* assertion of SYNC	*/
		ushort	syncend;	/* deassertion of SYNC	*/
		ushort	blankend;	/* deassertion of BLANK	*/
		ushort	total;		/* total pixels/lines	*/
		ushort	polarity;	/* sync polarity	*/
	};

	struct kgi_preset
	{
		int	dclk;		/* dot clock in Hz	*/
		struct kgi_timing x,y;	/* x and y timing values*/
	};

	struct kgi_mode
	{			/* mode setup info, calculated or preset*/
		ggi_mode	request;/* requested mode	*/

	  			/* calculated 'on the fly'	*/
	struct	kgi_timing	x,y;	/* timing parameters	*/
		ggi_coord	magnify;/* stretched ... times	*/
	enum	kgi_flags	flags;	/* special capabilities	*/
		int		dclk;	/* dot (master)		*/
		ggi_ratio	clk;	/* clock rate programmed*/
		ggi_ratio	lclk;	/* transfer rate to dac (byte/s)*/

					/* set by kernel driver	*/
		mem_vaddr	fb;	/* fb address		*/
		ulong		fb_size;/* total buffer size	*/
		ulong		fsize;	/* size of one frame	*/

		struct kgi_fb_ops *io;	/* ops to use when accessing hw */
		struct kgi_fb_ops *mem;	/* ops to use when accessing mem */
		struct kgi_mmio_region *mmio;	/* possible mmio regions */
		struct kgi_accel	*accel;	/* possible accelerators */

/* !!! change them to union driver_info ... */
		int		monitor;/* driver specific,	*/
		int		ramdac;	/* check_timing should	*/
		int		clock;	/* assume them not valid*/
		int		chipset;
	};

	enum kgi_tm_cmd
	{
		CMD_PROPOSE,		/* propose a timing	*/
		CMD_LOWER,		/* only lower clock rate*/
		CMD_RAISE,		/* only raise clock rate*/
		CMD_CHECK,		/* don't modify but check*/
		CMD_READY		/* don't care.		*/
	};

#define	INFOVERSION	01

	union kgi_info		/* union to ensure proper mem allocation  */
	{
		struct kgi_common	common;

		struct kgi_monitor	monitor;
		struct kgi_ramdac	ramdac;
		struct kgi_clock	clock;
		struct kgi_chipset	chipset;
	};

	struct kgi_card
	{
		KGI_COMMON
		struct kgi_monitor	*monitor;
		struct kgi_ramdac	*ramdac;
		struct kgi_clock	*clock;
		struct kgi_chipset	*chipset;

		struct kgi_mode		preset;
	};

/* ----------------------------------------------------------------------------
 *	display and input management
 * ----------------------------------------------------------------------------
 *	Some terms:
 *
 *	input	- device that converts real world actions into events
 *		- has a class (the general type, e.g. character, pointing ...)
 *		- has a type (e.g is a keyboard, a mouse, joystick, etc.)
 *		- has a focus (where the events are reported to)
 *
 *	keyboard- is a specialized input device for character and (optional)
 *		  pointer position input.
 *		- has up to 128 keys, 8 leds and a beeper to make some noise.
 *		- is the only input device that is allowed to change focus
 *
 *	display	- device that converts pixeldata into realworld images.
 *		- has to be enabled before any access using ENABLE_DISPLAY(dpy);
 *		- has a certain mode that determines how pixeldata is converted
 *		- may have up to eight framebuffers/CLUTs (one or two displayed)
 *		- may have (hardware) pointer/cursor shape.
 *
 *	device	- a combination of a display and input facilities (kbd, ptr,...)
 *		- is mapped (displayed) on a display
 *		- may be either a console or a graphic device
 *
 *	console	- special type of device that displays characters
 *		- always in text mode (may be graphic mode too later)
 *
 *	graphic	- special type of device (gives 'raw' mode access to a display)
 *		- gives 'raw' mode reporting of events
 */

#define	ENABLE	1
#define	DISABLE	0

#define display	kgi_display
#define	pixmap	kgi_pixmap
#define	pixel	ggi_pixel
#define	attr	ggi_attr
#define	fb_ops	kgi_fb_ops
#define	FBARGD	pixmap *fb
#define	FBARGC	fb

#include <ggi/pixmap.h>

#undef	display
#undef	pixmap
#undef	pixel
#undef	attr
#undef	fb_ops
#undef	FBARGD
#undef	FBARGC

#define	MAX_NR_DISPLAYS		32
#define	MAX_NR_DEVICES		(2*MAX_NR_CONSOLES)
#define	MAX_NR_FOCUSES		32
#define	NR_FB			8	/* (max) frames per display	*/

struct kgi_input
{
	struct kgi_input	*next;	/* next in linked list		*/
	ggi_sint	focus;			/* who gets events from us	*/
	ggi_uint	id;			/* input device ID		*/
	ggi_event_mask	events;		/* event types generated	*/
	ggi_event_mask	report;		/* events to report		*/

	char	manufact[64];		/* who built this device	*/
	char	model[64];		/* what model			*/

	int	(*ctrl)(struct kgi_input *, ggi_uint cmd, void *);
};

#define	KGI_IS_FOCUSED(i) 	((0 <= i.focus) && (i.focus < MAX_NR_DEVICES) \
				&& kgidev[i.focus])
#define	KGI_VALID_FOCUS(i)	((i.focus < 0) || ((0 <= i.focus) && \
				(i.focus < MAX_NR_DEVICES)))

#define	VALID_FOCUS_ID(x)	((0 <= x) && (x < MAX_NR_FOCUSES))
#define	VALID_DEV_ID(x)		((0 <= x) && (x < MAX_NR_DEVICES))
#define	VALID_DPY_ID(x)		((0 <= x) && (x < MAX_NR_DISPLAYS))

struct kgi_keymap
{
	ggi_uint	fn_buf_size;	/* size of function string buffer	*/
	char	*fn_buf;	/* pointer to funtion string buffer	*/

	ggi_uint	fn_str_size;	/* size of function key string buffer	*/
	char	**fn_str;	/* ptr to arr of ptrs to strings	*/

	ggi_uint	keymin, keymax;	/* min/max key values accepted		*/
	ggi_uint	keymap_size;	/* size of keymap array			*/
	unicode **keymap;	/* the keymaps				*/
};

struct kgi_focus
{
	ggi_uint	id;
	int	focus;
	struct kgi_input	*inputs;	/* input devices	*/

	MODIFIER_DATA;				/* modifier information	*/
	ggi_uint			down_mod[16];	/* modifier counters	*/
	struct kgi_keymap	*kmap;		/* key map info		*/

	ggi_uint	flags;
	#define	CAPS_SHIFT	1
	#define	HAS_POINTER	2

	ggi_sint	x, y;		/* pointer hot spot position	*/
	ggi_uint	buttons;	/* pointer button state		*/
	ggi_sint	maxx, maxy;	/* pointer hot spot window	*/
	ggi_sint	minx, miny;	/* pointer hot spot window	*/

	int	last_focus;
	int	want_focus;
	int	npadch;
	ggi_key_event dead;	/* pending dead/compose event	*/
};

extern int kbd_init(void);

struct kgi_mmio_region 
{
	struct kgi_mmio_region *next;
	char	*name;	/* name of the region		*/
	int	type;	/* standard types		*/
	#define MMIO_FRAMEBUFFER 1

	int	prot;   /* mapping protection flags	*/
	#define	MMIO_RD		VM_READ
	#define	MMIO_WR		VM_WRITE 
	#define	MMIO_RDWR	(VM_READ | VM_WRITE)
	#define	MMIO_SHARED	VM_SHARED

        ulong	size;  	/* total size of the region     */
	ulong	psize;	/* physical size of window	*/
        ulong	pstart; /* physical address of window   */
	ulong	offset;	/* logical address at pstart	*/

	void	(*set_offset)(struct kgi_mmio_region *, ggi_uint offset);
	void	*private;
};

/* !!! fix system dependent data types! */
struct kgi_mmio_mapping
{
	struct kgi_mmio_mapping	*next;	/* next in list			*/

	ggi_uint	devid;			/* device this belongs to	*/
	ggi_uint	mmap;			/* MMAP_LINEAR or MMAP_PAGED	*/
		#define	MAP_LINEAR_LINEAR	1
		#define	MAP_LINEAR_PAGED	2
		#define	MAP_PAGED_LINEAR	3
		#define	MAP_PAGED_PAGED		4

	struct kgi_mmio_region *r;	/* the region we are mapping	*/
	struct kgi_mmio_mapping **other;/* pointer to list head		*/
	struct file *file;

	ulong	offset;			/* offset for paged mappings	*/
	struct mm_struct *mm;		/* mapping table struct		*/
	ulong	mstart, mend;		/* start/end of mapped region	*/
};

struct kgi_accel_buffer {

	struct kgi_accel_buffer	*next;	/* next buffer to process	*/
	struct wait_queue	*ready;	/* wake up when becoming idle	*/
	ggi_uint	state;			/* state			*/
		#define	ACCEL_IDLE	0	/* nothing to do	*/
		#define	ACCEL_FILL	1	/* gets filled by app	*/
		#define	ACCEL_EXEC	2	/* being executed	*/
		#define	ACCEL_WAIT	3	/* waiting for exexution*/
	ggi_uint	size;			/* size of buffer		*/
	void	*buffer;		/* page aligned pointer		*/
	void	*alloc;			/* pointer to allocated memory	*/
};

struct kgi_accel {

	struct kgi_accel	*next;	/* next accelerator		*/
	struct kgi_display	*dpy;

	char	*name;			/* identifier			*/
	ggi_uint	type;			/* numeric ID			*/

	void	(*execute)(struct kgi_accel*, struct kgi_accel_buffer *);
};

struct kgi_accel_mapping {

	struct kgi_accel_mapping *next;	/* next mapping of this accel	*/
	struct kgi_accel_mapping **other;/*other mappings of this accel	*/
	struct kgi_accel	*accel;	/* the executing accelerator	*/
	
	struct file 	*file;		/* file this is associated to	*/
	ggi_uint		devid;		/* ID of special file		*/

	struct kgi_accel_buffer	ping;	/* command buffer		*/
	struct kgi_accel_buffer	pong;	/* command buffer		*/
};

struct kgi_graphic;

typedef int kgi_command_fn(struct kgi_graphic *, ggi_uint, ggi_uint);

enum kgi_pointer_mode
{

	pmWindows, pmX11, pmThreeColor
};

struct kgi_pointer64x64
{
	enum kgi_pointer_mode	mode;
	uint8	xor_mask[512];
	uint8	and_mask[512];
	ggi_uint	hotx, hoty;
	ggi_color	col[3];
};

struct kgi_display
{				/* these are set by the kgi kernel code	*/
	int	id;		/* display number, init to -1		*/
	ggi_uint	refcnt;		/* reference count, init to 0		*/
	ulong	conflict_text;	/* displays in conflict when in textmode*/
	ulong	conflict_graph;	/* displays in conflict when in graphics*/
	ulong	conflict_setup;	/* displays in conflict during setup	*/
	struct kgi_device *dev;	/* device currently mapped		*/

				/* these are set by the driver		*/
	void	*driver_data;		/* internal data of the driver	*/

	void	(*enable)(struct kgi_display *);
	void	(*disable)(struct kgi_display *);
	void	(*reset)(struct kgi_display *);

	kgi_command_fn *cmd[KGI_NR_SUBSYSTEMS];	/* subsystem command handlers */
	struct kgi_card		*card;		/* info about the hardware */

	int (*check_mode)(struct kgi_display*,struct kgi_mode*,enum kgi_tm_cmd);

        void (*set_mode)(struct kgi_display *, struct kgi_mode *);
	void (*set_font)(struct kgi_display *, struct kgi_font *, ggi_uint, ggi_uint);
	void (*set_clut)(struct kgi_display *, ggi_clut *, ggi_uint, ggi_uint);
	void (*set_split)(struct kgi_display*, int);
	void (*set_frame)(struct kgi_display*, int);
	void (*set_origin)(struct kgi_display *, ggi_uint, ggi_uint);
	void (*set_pointer_shape)(struct kgi_display*,struct kgi_pointer64x64*);

        void (*show_pointer)(struct kgi_display *, ggi_uint, ggi_uint);
	void (*hide_pointer)(struct kgi_display *);
        void (*undo_pointer)(struct kgi_display *);/*only if show modifies fb */

	void (*show_cursor)(struct kgi_display *, ggi_uint, ggi_uint);
	void (*hide_cursor)(struct kgi_display *);
	void (*undo_cursor)(struct kgi_display *); /*only if show modifies fb */

	struct kgi_mode	mode;		/* current mode of the display	*/

	ggi_mode	*defmode;	/* default mode for this display*/
	kgi_pixmap	fb;		/* framebuffer of current frame	*/
};

struct kgi_device	/* common fields for console/graphic devices	*/
{
	int	id;
	int	refcnt;

	ushort	type;			/* console or graphic device	*/
	#define	KGI_CONSOLE_DEV	0
	#define	KGI_GRAPHIC_DEV	1

	ushort	state;			/* various state control bits	*/
	#define	KGI_MAPPED_ON_DISPLAY	1	/* display map state	*/
	#define	KGI_DO_BACKINGSTORE	2	/* keep fb when in bg	*/
	#define	KGI_POINTER_TO_SHOW	4	/* show pointer		*/
	#define	KGI_POINTER_SHOWN	8	/* pointer visible	*/
	#define	KGI_POINTER_FB_MODIFIED	16	/* pointer drawn in fb	*/
	#define	KGI_CURSOR_TO_SHOW	32	/* show cursor		*/
	#define	KGI_CURSOR_SHOWN	64	/* cursor visible	*/
	#define	KGI_CURSOR_FB_MODIFIED	128	/* cursor drawn in fb	*/
	#define	KGI_DEVICE_IS_FOCUSED	256	/* is currently focused	*/

	/* set by KGI manager when the device is attached to a display */
	struct	kgi_display	*dpy;	/* display used for viewing	*/
	ggi_uint			focus;	/* focus that sends input	*/
	struct	kgi_fb_ops	*mem;	/* framebuffer ops when in bg	*/

		ggi_mode	mode;	/* display mode to be used	*/
		ggi_uint		frame;	/* frame displayed		*/
		ggi_uint		split;	/* split position for frame 0	*/
		ggi_coord	org[NR_FB]; /* origin *pixel*		*/
		ggi_clut	lut[NR_FB]; /* look up tables		*/
		kgi_pixmap	 fb[NR_FB]; /* frame buffers		*/
	struct	kgi_font	font;	/* font to use for text output	*/
	struct kgi_pointer64x64	*ptr_shape;	/* pointer shape	*/
		ggi_coord	cursor;		/* cursor hot spot	*/
		ggi_coord	pointer;	/* pointer hot spot	*/
		ggi_coord	ptr_max;	/* pointer window max	*/
		ggi_coord	ptr_min;	/* pointer window min	*/

	void (*save)(struct kgi_device *);
	void (*restore)(struct kgi_device *);
	void (*handle_event)(struct kgi_device *, ggi_event *);
	ggi_uint	event_mask;		/* selected events		*/

	struct	tty_struct	*tty;	/* associated tty		*/
};

#include <ggi/console.h>
#include <ggi/graphic.h>


#ifdef __KERNEL__

/*\
|*| ----------------------------------------------------------------------------
|*|	KGI manager interface
|*| ----------------------------------------------------------------------------
\*/

extern struct kgi_display	*kgidpy[MAX_NR_DISPLAYS];
extern struct kgi_focus		*kgifocus[MAX_NR_FOCUSES];
extern struct kgi_device	*kgidev[MAX_NR_DEVICES];

extern void *kgi_alloc_device(ggi_uint num, int priority);
extern void kgi_free_device(struct kgi_device *dev);
#define kgi_device_allocated(num) (device[num])

extern int kgi_attach_device(struct kgi_device *dev);
extern void kgi_detach_device(struct kgi_device *dev);
extern void kgi_map_device(struct kgi_device *dev);
extern void kgi_unmap_device(struct kgi_device *dev);
extern void kgi_show_gadgets(struct kgi_device *dev);
extern void kgi_undo_gadgets(struct kgi_device *dev);
extern void kgi_hide_gadgets(struct kgi_device *dev);

extern int kgi_check_mode(struct kgi_display*,struct kgi_mode*,enum kgi_tm_cmd);
extern void kgi_set_mode(struct kgi_display *, struct kgi_mode *);
extern void kgi_set_pointer(struct kgi_device *dev, ggi_uint x, ggi_uint y);
extern void kgi_set_pointer_window(struct kgi_device *dev, ggi_uint xmin, ggi_uint xmax,
	ggi_uint ymin, ggi_uint ymax);
extern void kgi_set_origin(struct kgi_device *dev, ggi_uint frame, ggi_uint x, ggi_uint y);
extern void kgi_set_split(struct kgi_device *dev, ggi_uint line);

extern void kgi_init(void);
extern ulong kgi_console_init(ulong kmem_start, ulong kmem_end);
extern ulong kgi_boot_init(ulong kmem_start, ulong kmem_end);
extern int display_init(void);

extern int kgi_register_display(struct kgi_display *dpy, int id);
extern int kgi_unregister_display(struct kgi_display *dpy, struct kgi_display *alias);
extern void kgi_enable_display(struct kgi_display *dpy, int setup);

extern int kgi_register_input(int focus, struct kgi_input *dev);
extern void kgi_unregister_input(int focus, struct kgi_input *dev);

extern void kgi_focus_device(ggi_uint dev);
extern void kgi_handle_input(ggi_event *);
extern char *kgi_function_key_string(int focus, ggi_uint fnkey);
extern unicode kgi_combine_dead(int focus, ggi_uint dead, ggi_uint key);

extern ggi_clut *kgi_default_clut(ggi_mode *mode);

extern kgi_fb_ops	text16_mono_io, text16_mono_mem;
extern kgi_fb_ops	text16_color_io, text16_color_mem;

extern struct kgi_display *text16_malloc_display(size_t fb, ggi_uint x, ggi_uint y,
	ggi_uint textx, ggi_uint texty, int color);
extern void text16_free_display(struct kgi_display *dpy);

extern struct kgi_focus default_focus;

#endif	/* #ifdef __KERNEL__ */

#endif	/* #ifdef _GGI_KGI_H */
